import time
from multiprocessing import Event, Queue
from Queue import Empty
import errno
import socket

from virtwho import log, MinimumSendInterval, MinimumJobPollInterval

from virtwho.config import ConfigManager
from virtwho.manager import (
    Manager, ManagerThrottleError, ManagerError, ManagerFatalError)
from virtwho.virt import (
    AbstractVirtReport, ErrorReport, DomainListReport,
    HostGuestAssociationReport, Virt)

try:
    from collections import OrderedDict
except ImportError:
    # Python 2.6 doesn't have OrderedDict, we need to have our own
    from util import OrderedDict


class ReloadRequest(Exception):
    ''' Reload of virt-who was requested by sending SIGHUP signal. '''


class Executor(object):
    def __init__(self, logger, options, config_dir=None):
        """
        Executor class provides bridge between virtualization supervisor and
        Subscription Manager.

        logger - logger instance
        options - options for virt-who, parsed from command line arguments
        """
        self.logger = logger
        self.options = options
        self.terminate_event = Event()
        self.virts = []

        # Queue for getting events from virt backends
        self.queue = None

        # Dictionary with mapping between config names and report hashes,
        # used for checking if the report changed from last time
        self.last_reports_hash = {}
        # How long should we wait between reports sent to server
        self.retry_after = MinimumSendInterval
        # This counts the number of responses of http code 429
        # received between successfully sent reports
        self._429_count = 0
        self.reloading = False
        # The last time a 429 was returned from the server
        self.last_429_time = 0

        # Reports that are queued for sending
        self.queued_reports = OrderedDict()

        # Name of configs that wasn't reported in oneshot mode
        self.oneshot_remaining = set()

        # Report that is currently being processed by server
        self.reports_in_progress = OrderedDict()

        # True if this is the first time we are checking job status for a given job
        self.is_first_job_check = True

        # Next time to check the status of the most recently submitted job
        self.next_job_status_check = 0

        self.configManager = ConfigManager(self.logger, config_dir)

        for config in self.configManager.configs:
            logger.debug("Using config named '%s'" % config.name)

        self.send_after = time.time()

    def check_report_state(self, report):
        ''' Check state of one report that is being processed on server. '''
        manager = Manager.fromOptions(self.logger, self.options, report.config)
        manager.check_report_state(report)
        if report.state == AbstractVirtReport.STATE_CREATED:
            self.logger.warning("Can't check status of report that is not yet sent")
            self.next_job_status_check = time.time() + MinimumJobPollInterval
        elif report.state != AbstractVirtReport.STATE_PROCESSING:
            # Any other state we consider to be done
            self.report_done(report)
        return report.state

    def _is_next_job_check_time(self):
        # If there are no jobs to check, it's not time to check them
        if not self.reports_in_progress:
            return False
        return time.time() >= self.next_job_status_check

    def send_current_report(self):
        name, report = self.queued_reports.popitem(last=False)
        if self.last_reports_hash.get(name, None) == report.hash:
            return
        return self.send_report(name, report)

    def _handle_429(self, retry_after):
        self._429_count += 1
        self.retry_after = retry_after
        self.last_429_time = time.time()
        self.logger.debug('429 received, waiting %s seconds until sending again', self.retry_after)

    def _reset_429_handler(self):
        self._429_count = 0
        self.retry_after = 0
        self.last_429_time = 0

    def send_report(self, name, report):
        try:
            if self.send(report):
                self.logger.debug('Report for config "%s" sent', name)
                if report.state == AbstractVirtReport.STATE_PROCESSING:
                    if not self.reports_in_progress:
                        self.next_job_status_check = time.time() + MinimumJobPollInterval
                    self.reports_in_progress[report.config.name] = report

                else:
                    self.report_done(report)
            else:
                report.state = AbstractVirtReport.STATE_FAILED
                self.logger.debug('Report from "%s" failed to sent', name)
                self.report_done(report)
        except ManagerThrottleError as e:
            self.queued_reports[name] = report
            self._handle_429(max(MinimumSendInterval, e.retry_after))
            self.send_after = time.time() + self.retry_after

    def report_done(self, report):
        name = report.config.name
        self.send_after = time.time() + self.options.interval
        if report.state == AbstractVirtReport.STATE_FINISHED:
            self.last_reports_hash[name] = report.hash

        if report.state in [AbstractVirtReport.STATE_FINISHED,
                            AbstractVirtReport.STATE_FAILED,
                            AbstractVirtReport.STATE_CANCELED]:
            if name in self.reports_in_progress:
                del self.reports_in_progress[name]
            if self.reports_in_progress:
                self.next_job_status_check = time.time() + MinimumJobPollInterval
            self.is_first_job_check = True

        if self.options.oneshot:
            self.oneshot_remaining.discard(name)

    def send(self, report):
        """
        Send list of uuids to subscription manager

        return - True if sending is successful, False otherwise
        """
        try:
            if isinstance(report, DomainListReport):
                self._sendGuestList(report)
            elif isinstance(report, HostGuestAssociationReport):
                self._sendGuestAssociation(report)
            else:
                self.logger.warn("Unable to handle report of type: %s", type(report))
        except ManagerError as e:
            self.logger.error("Unable to send data: %s", str(e))
            return False
        except ManagerFatalError:
            raise
        except ManagerThrottleError:
            raise
        except socket.error as e:
            if e.errno == errno.EINTR:
                self.logger.debug("Communication with subscription manager interrupted")
            return False
        except Exception as e:
            if self.reloading:
                # We want to skip error reporting when reloading,
                # it is caused by interrupted syscall
                self.logger.debug("Communication with subscription manager interrupted")
                return False
            exceptionCheck(e)
            self.logger.exception("Error in communication with subscription manager:")
            return False
        return True

    def _sendGuestList(self, report):
        manager = Manager.fromOptions(self.logger, self.options, report.config)
        manager.sendVirtGuests(report, self.options)

    def _sendGuestAssociation(self, report):
        manager = Manager.fromOptions(self.logger, self.options, report.config)
        manager.hypervisorCheckIn(report, self.options)

    def run(self):
        self.reloading = False
        if not self.options.oneshot:
            self.logger.debug("Starting infinite loop with %d seconds interval", self.options.interval)

        # Queue for getting events from virt backends
        if self.queue is None:
            self.queue = Queue()

        # Run the virtualization backends
        self.virts = []
        for config in self.configManager.configs:
            try:
                logger = log.getLogger(config=config)
                virt = Virt.fromConfig(logger, config)
            except Exception as e:
                self.logger.error('Unable to use configuration "%s": %s', config.name, str(e))
                continue
            # Run the process
            virt.start(self.queue, self.terminate_event, self.options.interval, self.options.oneshot)
            self.virts.append(virt)

        # This set is used both for oneshot mode and to bypass rate-limit
        # when virt-who is starting
        self.oneshot_remaining = set(virt.config.name for virt in self.virts)

        if len(self.virts) == 0:
            err = "virt-who can't be started: no suitable virt backend found"
            self.logger.error(err)
            exit(1, err)

        # queued reports depend on OrderedDict feature that if key exists
        # when setting an item, it will remain in the same order
        self.queued_reports.clear()

        # Clear last reports, we need to resend them when reloaded
        self.last_reports_hash.clear()

        # report that is being processed by the server
        self.reports_in_progress.clear()

        # Send the first report immediately
        self.send_after = time.time()

        # Time to check job status
        self.next_job_status_check = 0

        self.is_first_job_check = True

        while not self.terminate_event.is_set():
            # Block for the amount of time closest to the next action we need to perform
            # OR do not block at all if it is time to do one of them
            timeout = max(self.retry_after - (time.time() - self.last_429_time),
                          min(self.next_job_status_check - time.time(), self.send_after - time.time()))
            if (self.oneshot_remaining and self._429_count < 1) or timeout < 0:
                timeout = 1
            # Wait for incoming report from virt backend or for timeout
            try:
                report = self.queue.get(block=True, timeout=timeout)
            except Empty:
                report = None
            except IOError:
                continue

            # Read rest of the reports from the queue in order to remove
            # obsoleted reports from same virt
            while True:
                if isinstance(report, ErrorReport):
                    if self.options.oneshot:
                        # Don't hang on the failed backend
                        self.oneshot_remaining.discard(report.config.name)
                        self.logger.warn('Unable to collect report for config "%s"', report.config.name)
                elif isinstance(report, AbstractVirtReport):
                    if self.last_reports_hash.get(report.config.name, None) != report.hash:
                        self.queued_reports[report.config.name] = report
                        if report.config.name in self.oneshot_remaining:
                            if not self.options.print_:
                                # Report is not queued if it sent immediately
                                del self.queued_reports[report.config.name]
                                self.send_report(report.config.name, report)
                            self.oneshot_remaining.discard(report.config.name)
                    else:
                        self.logger.info('Report for config "%s" hasn\'t changed, not sending', report.config.name)
                elif report in ['exit', 'reload']:
                    # Reload and exit reports takes priority, do not process
                    # any other reports
                    break

                # Get next report from queue
                try:
                    report = self.queue.get(block=False)
                except Empty:
                    break

            if report == 'exit':
                break
            elif report == 'reload':
                self.stop_virts()
                raise ReloadRequest()

            if time.time() < (self.last_429_time + self.retry_after):
                continue
            else:
                self._reset_429_handler()

            if self._is_next_job_check_time():
                try:
                    report_state = self.check_report_state(self.reports_in_progress.items()[0][1])
                except ManagerThrottleError as e:
                    self.logger.debug('HTTP 429 received during job polling')
                    retry_after = e.retry_after if e.retry_after != 0 else (MinimumJobPollInterval * 2)
                    self.next_job_status_check = time.time() + retry_after
                    self._handle_429(retry_after)
                else:
                    self._reset_429_handler()
                    if report_state not in [AbstractVirtReport.STATE_CANCELED,
                                            AbstractVirtReport.STATE_FAILED,
                                            AbstractVirtReport.STATE_FINISHED]:
                        self.is_first_job_check = False
                        self.next_job_status_check = time.time() + (MinimumJobPollInterval * 2)

            if not self.reports_in_progress and self.queued_reports and time.time() > self.send_after:
                # No report is processed, send next one
                if not self.options.print_:
                    self.send_current_report()
            if self.options.oneshot and not self.oneshot_remaining and not self.reports_in_progress and (not self.queued_reports or self.options.print_):
                break

        self.queue = None
        self.stop_virts()

        self.virt = []
        if self.options.print_:
            return self.queued_reports

    def stop_virts(self):
        for virt in self.virts:
            virt.stop()
            virt.terminate()
            virt.join()
        self.virts = []

    def terminate(self):
        self.logger.debug("virt-who is shutting down")

        # Terminate the backends before clearing the queue, the queue must be empty
        # to end a child process, otherwise it will be stuck in queue.put()
        self.terminate_event.set()
        # Give backends some time to terminate properly
        time.sleep(0.5)

        if self.queue:
            # clear the queue and put "exit" there
            try:
                while True:
                    self.queue.get(False)
            except Empty:
                pass
            self.queue.put("exit")

        # Give backends some more time to terminate properly
        time.sleep(0.5)

        self.stop_virts()

    def reload(self):
        self.logger.warn("virt-who reload")
        # Set the terminate event in all the virts
        for virt in self.virts:
            virt.stop()
        # clear the queue and put "reload" there
        try:
            while True:
                self.queue.get(False)
        except Empty:
            pass
        self.reloading = True
        self.queue.put("reload")


def exceptionCheck(e):
    try:
        # This happens when connection to server is interrupted (CTRL+C or signal)
        if e.args[0] == errno.EALREADY:
            exit(0)
    except Exception:
        pass
